package de.mattv.fileserver.data;

import com.thoughtworks.xstream.io.HierarchicalStreamDriver;
import com.thoughtworks.xstream.io.xml.XppDriver;
import de.mattv.fileserver.Utils;
import de.mattv.fileserver.data.converter.ConverterUtils;
import de.mattv.fileserver.data.converter.DataConverter;
import jakarta.annotation.Nullable;
import jakarta.annotation.PostConstruct;
import jakarta.annotation.PreDestroy;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.time.Duration;
import java.time.Instant;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.locks.ReentrantReadWriteLock;

@Slf4j
@Component
public class Data {
    private static final File DATA_FILE = new File("data.xml");
    private static final File DATA_FILE_NEW = new File("data_new.xml");
    private static final File DATA_FILE_OLD = new File("data_old.xml");
    private static final HierarchicalStreamDriver X_STREAM = new XppDriver();

    private static Thread saveThread = null;
    private static final AtomicBoolean SAVE_FLAG = new AtomicBoolean(false);
    private static final AtomicBoolean SHUTDOWN_FLAG = new AtomicBoolean(false);

    public static final File FILES_DIR = new File("files");
    public static final long CURRENT_VERSION = 1;
    public static long nextUserId = 0;

    public static final ConcurrentHashMap<Long, User> USERS = new ConcurrentHashMap<>();


    public static final ReentrantReadWriteLock USER_LOCK = new ReentrantReadWriteLock();

    private static void saveData() {
        Instant now = Instant.now();
        log.info("Saving data");
        try (FileWriter writer = new FileWriter(DATA_FILE_NEW)) {
            DataConverter.toXml(X_STREAM.createWriter(writer));
        } catch (IOException e) {
            Utils.crash(log, "Failed to serialize data to xml", e);
        }
        try {
            if (DATA_FILE.exists())
                Files.move(Data.DATA_FILE.toPath(), Data.DATA_FILE_OLD.toPath(), StandardCopyOption.REPLACE_EXISTING);
            Files.move(Data.DATA_FILE_NEW.toPath(), Data.DATA_FILE.toPath(), StandardCopyOption.REPLACE_EXISTING);
        } catch (IOException e) {
            Utils.crash(log, "Failed move data xml", e);
        }
        double elapsed = Duration.between(now, Instant.now()).toNanos() / 1000000.0;
        log.info("Saved data in {} ms", elapsed);
    }

    public static void save() {
        SAVE_FLAG.set(true);
        synchronized (SAVE_FLAG) { SAVE_FLAG.notifyAll(); }
    }

    @PreDestroy
    public static void onExit() {
        SHUTDOWN_FLAG.set(true);
        save();
        try {
            saveThread.join();
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    @PostConstruct
    public static void load() {
        log.info("Loading data from {}", DATA_FILE.getAbsolutePath());
        try {
            if (!FILES_DIR.exists()) if (!FILES_DIR.mkdir()) Utils.crash(log, "Failed to create files dir");
            if(DATA_FILE.exists()) DataConverter.fromXML(X_STREAM.createReader(DATA_FILE));
            if (ConverterUtils.hasError) throw new RuntimeException();
            DataValidator.validateData();
        } catch (RuntimeException e) {
            Utils.crash(log, "Failed to load data", e);
        }
        log.info("Finished loading data");
        saveThread = new Thread(() -> {
            while (!SHUTDOWN_FLAG.get()) {
                try { synchronized (SAVE_FLAG) { SAVE_FLAG.wait(); } } catch (InterruptedException ignored) {}
                do {
                    SAVE_FLAG.set(false);
                    try {
                        //noinspection BusyWait
                        Thread.sleep(2000);
                    } catch (InterruptedException ignored) {}
                } while (SAVE_FLAG.get());
                saveData();
            }
            log.info("Data saver stopping");
            saveData();
        });
        saveThread.start();
    }

    public static @Nullable User findUser(String name) {
        return Data.USERS.values()
                .stream()
                .filter(u -> u.name.equals(name))
                .findFirst()
                .orElse(null);
    }
}
