package de.mattv.fileserver.data.converter;

import com.thoughtworks.xstream.io.HierarchicalStreamReader;
import com.thoughtworks.xstream.io.HierarchicalStreamWriter;
import de.mattv.fileserver.data.Node;
import de.mattv.fileserver.data.User;
import de.mattv.fileserver.util.AutoCloseLock;
import de.mattv.fileserver.util.NonNull;
import lombok.extern.slf4j.Slf4j;

@Slf4j
public class UserConverter {
    private static final String NODE_NAME = "user";
    private static final String ATTR_ID_NAME = "id";
    private static final String ATTR_NAME_NAME = "name";
    private static final String ATTR_PASSWORD_NAME = "password";
    private static final String ATTR_FLAGS_NAME = "flags";
    private static final String ATTR_NEXT_NODE_NAME = "next-node";
    private static final String ATTR_TFA_NAME = "tfa";

    private static class Flags {
        private static final long ENABLED = 1;
        private static final long ADMIN = 2;
    }

    public static void toXml(@NonNull User user, @NonNull HierarchicalStreamWriter writer) {
        try (AutoCloseLock ignored = new AutoCloseLock(user.nodeLock.readLock())) {
            writer.startNode(NODE_NAME);

            long flags = 0;
            if (user.enabled) flags |= Flags.ENABLED;
            if (user.admin) flags |= Flags.ADMIN;

            writer.addAttribute(ATTR_ID_NAME, String.valueOf(user.id));
            writer.addAttribute(ATTR_NAME_NAME, user.name);
            writer.addAttribute(ATTR_PASSWORD_NAME, user.password);
            writer.addAttribute(ATTR_FLAGS_NAME, String.valueOf(flags));
            writer.addAttribute(ATTR_NEXT_NODE_NAME, String.valueOf(user.nextNodeId));

            String tfa = user.tfaSecret;
            if (tfa != null) writer.addAttribute(ATTR_TFA_NAME, tfa);

            user.nodes.get(0L).children.forEach(child -> NodeConverter.toXml(child, writer));

            writer.endNode();
        }
    }

    public static User fromXml(@NonNull HierarchicalStreamReader reader) {
        reader.moveDown();

        if (!reader.getNodeName().equals(NODE_NAME)) {
            log.error("Trying to parse user from xml type {}", reader.getNodeName());
            ConverterUtils.dumpConversionStackFrames();
        }

        ConverterUtils.push("User");

        User user = new User(
                ConverterUtils.getLong(reader, ATTR_ID_NAME),
                ConverterUtils.getString(reader, ATTR_NAME_NAME),
                ConverterUtils.getString(reader, ATTR_PASSWORD_NAME)
        );

        ConverterUtils.pop();
        ConverterUtils.push("User " + user.id);

        long flags = ConverterUtils.getLong(reader, ATTR_FLAGS_NAME);
        user.enabled = (flags & Flags.ENABLED) != 0;
        user.admin = (flags & Flags.ADMIN) != 0;
        user.nextNodeId.set(ConverterUtils.getLong(reader, ATTR_NEXT_NODE_NAME));

        String tfa = reader.getAttribute(ATTR_TFA_NAME);
        if (tfa != null) {
            user.tfaSecret = tfa;
        }

        while (reader.hasMoreChildren()) {
            Node child = NodeConverter.fromXml(reader, user);
            child.parent = user.rootNode;
            user.rootNode.children.add(child);
        }

        ConverterUtils.pop();
        reader.moveUp();

        return user;
    }
}
